/*
    POSIX Validator of ANSI library functions
    This rule-file uses the list of approved posix.1 functions that are allowed from 
    ANSI header's per page 181, figure 8.2 "Posix.1" standard book by Fred Zlotnick.
    If a function is called and was declared in a ansi header, and is not on the list
    approved by zlotnick then a warning is emmitted. 

    Issues addressed besides the above.

    1.) Macro expansion names [ ctype.h ] isprint(), may actually CALL _isprint()
    2.) Function call's NOT declared by any header are assumed illegal.
    3.) Prior to all #include<>'s the _POSIX_SOURCE macro must be defined.
    4.) No macro's should be defined with leading underscores, per posix.1
*/

#include <check.cch>	/* standard CodeCheck header file for constants */
#include <ccdebug.cch>  // debug functions for writing rule-files

int posix_need;		/* 1 if _POSIX_SOURCE has not been defined */

/* system header files for both POSIX.1 and ANSI C */
int errno_include,  	/* 1 if header errno.h has been #included */
    limits_include,     /* 1 if header limits.h has been #included */
    setjmp_include,     /* 1 if header setjmp.h has been #included */
    signal_include;	    /* 1 if header signal.h has been #included */

/* system header files for POSIX.1 Only */
int dirent_include,	 /* 1 if dirent.h has been #included */
    fcntl_include,	 /* 1 if fcntl.h has been #included */
    grp_include,	 /* 1 if grp.h has been #included */
    pwd_include,	 /* 1 if pwd.h has been #included */
    sys_stat_include,	 /* 1 if sys/stat.h has been #included */
    sys_times_include,	 /* 1 if sys/times.h has been #included */
    sys_types_include,	 /* 1 if sys/types_include.h has been #included */
    sys_utsname_include, /* 1 if sys/utsname.h has been #included */
    sys_wait_include,    /* 1 if sys/wait.h has been #included */
    tar_include,	 /* 1 if tar.h has been #included */
    termios_include,	 /* 1 if termios.h has been #included */
    unistd_include,      /* 1 if unistd.h has been #included */
    utime_include;       /* 1 if utime.h has been #included */

/* system header files for ANSI C Only */
int assert_include,      /* 1 if assert.h has been #included */
    ctype_include,       /* 1 if ctype.h has been #included */
    float_include,       /* 1 if float.h has been #included */
    flow_include,        /* 1 if flow.h has been #included */
    locale_include,      /* 1 if locale.h has been #included */
    math_include,        /* 1 if math.h has been #included */
    stdarg_include,      /* 1 if stdarg.h has been #included */
    stddef_include,      /* 1 if stddef.h has been #included */
    stdio_include,       /* 1 if stdio.h has been #included */
    stdlib_include,      /* 1 if stdlib.h has been #included */
    string_include,      /* 1 if string.h has been #included */
    time_include;        /* 1 if time.h has been #included */

if (mod_begin)
{
  posix_need = 1;       // signal to look for macro _POSIX_SOURCE

/* 
Make sure that project headers are checked, and that system headers are not checked 
(option -S1) [ default is -S0 .c / .cpp only ]
*/		

// comment out next line if #include "" checking is NOT required.

  set_option('S', 1);					 
}

if (pp_macro)
{
  if (prefix("_"))
  {
    if (strcmp(pp_name(), "_POSIX_SOURCE") == 0)
      posix_need = 0;
    else
      warn(2001, "%s: Leading underscores are reserved by POSIX.",
           dcl_name());
  }
}
if (header_name())	/* triggers whenever an #include is met. */
{
  if (posix_need)
  {
    warn(2000, "Precede all headers with #define _POSIX_SOURCE");
    posix_need = 0;	/* prevent multiple warning */
  }

  /* system headers for both POSIX.1 and ANSI C */
  if (strcmp(header_name(), "errno.h") == 0)
    errno_include = 1;
  else if (strcmp(header_name(), "limits.h") == 0)
    limits_include = 1;
  else if (strcmp(header_name(), "setjmp.h") == 0)
    setjmp_include = 1;
  else if (strcmp(header_name(), "signal.h") == 0)
    signal_include = 1;

  /* system headers for POSIX.1 Only */
  else if (strcmp(header_name(), "dirent.h") == 0)
    dirent_include = 1;
  else if (strcmp(header_name(), "fcntl.h") == 0)
    fcntl_include = 1;
  else if (strcmp(header_name(), "grp.h") == 0)
    grp_include = 1;
  else if (strcmp(header_name(), "pwd.h") == 0)
    pwd_include = 1;
  else if (strcmp(header_name(), "sys/stat.h") == 0)
    sys_stat_include = 1;
  else if (strcmp(header_name(), "sys/times.h") == 0)
    sys_times_include = 1;
  else if (strcmp(header_name(), "sys/types.h") == 0)
     sys_types_include = 1;
  else if (strcmp(header_name(), "sys/utsname.h") == 0)
    sys_utsname_include = 1;
   else if (strcmp(header_name(), "sys/wait.h") == 0)
    sys_wait_include = 1;
  else if (strcmp(header_name(), "tar.h") == 0)
    tar_include = 1;
  else if (strcmp(header_name(), "termios.h") == 0)
    termios_include = 1;
  else if (strcmp(header_name(), "unistd.h") == 0)
    unistd_include = 1;
  else if (strcmp(header_name(), "utime.h") == 0)
    utime_include = 1;

  /* system headers for ANSI C Only */
  else if (strcmp(header_name(), "assert.h") == 0)
    assert_include = 1;
  else if (strcmp(header_name(), "ctype.h") == 0)
    ctype_include = 1;
  else if (strcmp(header_name(), "float.h") == 0)
    float_include = 1;
  else if (strcmp(header_name(), "flow.h") == 0)
    flow_include = 1;
  else if (strcmp(header_name(), "locale.h") == 0)
    locale_include = 1;
  else if (strcmp(header_name(), "math.h") == 0)
    math_include = 1;
  else if (strcmp(header_name(), "stdarg.h") == 0)
    stdarg_include = 1;
  else if (strcmp(header_name(), "stddef.h") == 0)
    stddef_include = 1;
  else if (strcmp(header_name(), "stdio.h") == 0)
    stdio_include = 1;
  else if (strcmp(header_name(), "stdlib.h") == 0)
    stdlib_include = 1;
  else if (strcmp(header_name(), "string.h") == 0)
    string_include = 1;
  else if (strcmp(header_name(), "time.h") == 0)
    time_include = 1;
}

#define IDLEN   132
// This code act's like a function in order to validate if a current identifier is ANSI
int include_ansi;
int check_ansi;     // user trigger to act like function
int posix_ok;        // function was ansi and posix compliant
char IDNAME[ IDLEN ];

// logic for macro expansion exception [ ctype.h isprint() -> _isprint() ]
int macro_call;
//char MAC_ALIAS[ IDLEN ];

#define POK(s) { if( strcmp(IDNAME,s)==0 ) posix_ok=lin_number; }

// The following user-trigger is ugly, but is designed for easy addition of new approved
// functions to the list.

if ( check_ansi ) {

/* system headers for ANSI C Only */
  if (strcmp(idn_filename(), "assert.h") == 0) 
  {
      POK("assert")                 // example: assert() from #include<assert.h> is OK
                                    // any other function from assert.h will issue warning
      include_ansi = 1;           
  }
  else if (strcmp(idn_filename(), "ctype.h") == 0)
    {       
        POK("islanum")
        POK("isalpha")
        POK("iscntrl")
        POK("isdigit")
        POK("isgraph")

        POK("islower")
        POK("isprint")
        POK("ispunct")
        POK("isspace")
        POK("isupper")

        POK("isxdigit")
        POK("tolower")
        POK("toupper")

        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "float.h") == 0)
    {       
        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "flow.h") == 0)
    {       
        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "locale.h") == 0)
    {       
        POK("setlocale")

        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "math.h") == 0)
    {       
        POK("acos")
        POK("asin")
        POK("atan")
        POK("atans")
        POK("ceil")

        POK("cos")
        POK("cosh")
        POK("exp")
        POK("fabs")
        POK("floor")

        POK("fmod")
        POK("frexp")
        POK("ldexp")
        POK("log")
        POK("log10")

        POK("modf")
        POK("pow")
        POK("sin")
        POK("sinh")
        POK("sqrt")

        POK("sqrt")
        POK("tan")
        POK("tanh")

        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "setjmp.h") == 0)
    {      
        POK("setjmp")
        POK("longjmp")

        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "stdarg.h") == 0)
    {       
        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "stddef.h") == 0)
    {       
        include_ansi = 1;          
    }
  else if (strcmp(idn_filename(), "stdio.h") == 0)
    {    
        POK("clearerr")
        POK("fclose")
        POK("feof")
        POK("ferror")
        POK("fflush")

        POK("fgetc")
        POK("fgets")
        POK("fopen")
        POK("fprintf")
        POK("fputc")

        POK("puts")
        POK("fread")
        POK("freopen")
        POK("fscanf")
        POK("fseek")

        POK("ftell")
        POK("fwrite")
        POK("getc")
        POK("getchar")
        POK("gets")

        POK("perror")
        POK("printf")
        POK("putc")
        POK("putchar")
        POK("puts")

        POK("remove")
        POK("rename")
        POK("rewind")
        POK("scanf")
        POK("sscanf")

        POK("setbuf")
        POK("sprintf")
        POK("tmpfile")
        POK("tmpname")
        POK("ungetc")

        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "stdlib.h") == 0)
    {       
        POK("abort")
        POK("abs")
        POK("atof")
        POK("atoi")
        POK("atol")

        POK("bsearch")
        POK("calloc")
        POK("exit")
        POK("free")
        POK("getenv")

        POK("malloc")
        POK("qsort")
        POK("rand")
        POK("realloc")
        POK("srand")

        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "string.h") == 0)
    {       
        POK("strcat")
        POK("strchr")
        POK("strcmp")
        POK("strcpy")
        POK("strcspn")

        POK("strlen")
        POK("strncat")
        POK("strncmp")
        POK("strncpy")
        POK("strpbrk")

        POK("strrchr")
        POK("strspn")
        POK("strstr")
        POK("strtok")

        include_ansi = 1;           
    }
  else if (strcmp(idn_filename(), "time.h") == 0)
    {       

        POK("asctime")
        POK("ctime")
        POK("grntime")
        POK("loclatime")
        POK("mktime")

        POK("strftime")
        POK("time")

        include_ansi = 1;           
    }
}


if ( op_macro_call ) {
//warn(2, "OMC %s OM %s OF %s",token(), op_macro(), op_function() );
    if ( macro_call == 0 ) {             // only do once, for primary expansion  
        macro_call = lin_number;         // record expansion line
        strcpy( IDNAME, op_function() ); // save original name
    }
}



if ( idn_function ) {

//warn( 0, "%s() %d %s", idn_name(), idn_fileline(), idn_filename() );
    
    if ( idn_fileline() ) {                 // header name of where fun() declared

        if ( strstr(idn_filename(),".h") ) {

            if ( lex_macro_token == 0 ) {
                strcpy( IDNAME, idn_name() );
            }
              
            check_ansi = 1;                 // run/fire ansi-posix validator
            macro_call = 0;                 // clear
        }
//  warn( 1, "IF IDN=%s idn=%s() IA=%d PO=%d LMT=%d", IDNAME,idn_name(), include_ansi, posix_ok, lex_macro_token );

        if ( include_ansi && posix_ok != lin_number ) {
            warn( 200, "POSIX: ANSI Function Not Allowed %s() [%s]", idn_name(), idn_filename() );
        }

        include_ansi = 0;
    }
    else {                                  // no header-name, undeclared function.

        warn( 100, "Unknown Function Call %s() found", idn_name() );
    }
    
    
}